# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Plots Plugin API.
"""

from spyder.plugins.plots.widgets.main_widget import (
    PlotsWidgetActions, PlotsWidgetMainToolbarSections)
